/* allocate.cc
 $Header: /al/acs/src/RCS/allocate.cc,v 8.1 95/04/23 18:19:20 al Exp $
 * Allocates space for the admittance matrix.  Allocation below the
 * diagonal is by row, above the diagonal is by column, and stored backwards.
 * This broken vector is stored.  The length of it increases with increasing
 * position.  The maximum length of the nth vector is 2n-1.  For a band matrix
 * only those elements that are non-zero or are nearer to the diagonal than a
 * non-zero element are stored.
 */
#include "ecah.h"
#include "e_node.h"
#include "e_card.h"
#include "error.h"
#include "mode.h"
#include "u_option.h"
#include "u_status.h"
#include "m_matrix.h"
#include "declare.h"
/*--------------------------------------------------------------------------*/
	void	 allocate(int);
static	void	 count_nodes();
static	void	 map_nodes();
static	void	 order_reverse();
static	void	 order_forward();
static	void	 order_auto();
static	void	 map_list(CARD*);
static	void	 addto_basenode_table(const CARD*);
static	void	 alloc_hold_vectors();
static	void	 alloc_constants();
/*--------------------------------------------------------------------------*/
extern struct nodestuff ns;
extern BSMATRIXdouble aa, lu;
extern BSMATRIXcomplex acx;
extern const char e_om[];
extern struct nodestat *nstat;	    /* node status flags		    */
/*--------------------------------------------------------------------------*/
/* allocate: memory allocation for simulation
 */
void allocate(int mode)
{
 if (mode != sSTATUS){
    STATUS::load.reset();
    STATUS::lud.reset();
    STATUS::back.reset();
    STATUS::review.reset();
    STATUS::output.reset();
    STATUS::setup.zstart();
 }
 if (!nstat){		    	    	/* if "nstat" not allocated,	    */
    dealloc(YES);		    	/* must allocate everything	    */
    count_nodes();		 	/* and start solution over.	    */
    CARD::expand_all();
    map_nodes();
    alloc_hold_vectors();
    aa.reinit(STATUS::total_nodes);
    lu.reinit(STATUS::total_nodes);
    acx.reinit(STATUS::total_nodes);
    addto_basenode_table(CARD::first());
 }
 if (mode != sSTATUS){
    if (!exists(CARD::first()))
      error(bERROR, "no circuit\n");
    alloc_constants();
    STATUS::setup.stop();
 }
}
/*--------------------------------------------------------------------------*/
/* count_nodes: count nodes in main ckt (not subckts)
 * update the variables "STATUS::total_nodes" and "STATUS::user_nodes"
 * zeros "STATUS::subckt_nodes" and "STATUS::model_nodes"
 */
static void count_nodes()
{
 const CARD *stop;
 const CARD *brh;

 STATUS::user_nodes = 0;
 brh = stop = CARD::first();
 assert(exists(stop));
 do {
    if (brh->isdevice()){
       for (int ii = 0;  brh->n[ii].e != INVALIDNODE;  ii++){
	  if (brh->n[ii].e > STATUS::user_nodes)
	     STATUS::user_nodes = brh->n[ii].e;
       }
    }
 } while (brh = brh->next(),  brh != stop);
 STATUS::total_nodes = STATUS::user_nodes;
 STATUS::subckt_nodes = STATUS::model_nodes = 0;
}
/*--------------------------------------------------------------------------*/
/* map_nodes: map intermediate node number to internal node number.
 * Ideally, this function would find some near-optimal order
 * and squash out gaps.
 */
static void map_nodes()
{
 ns.nm = (int*)calloc((unsigned)STATUS::total_nodes+1, sizeof(int));
 if (!ns.nm)				/* node map table		    */
    error(bERROR, e_om, "nm");
 STATUS::order.zstart();
 switch (OPT::order){
    default:
       error(bWARNING, "invalid order spec: %d\n", OPT::order);
       /* fall through to order_reverse. */
    case oAUTO:
       order_auto();
       break;
    case oREVERSE:
       order_reverse();
       break;
    case oFORWARD:
       order_forward();
       break;
 }
 STATUS::order.stop();
 map_list(CARD::first());
}
/*--------------------------------------------------------------------------*/
/* order_reverse: force ordering to reverse of user ordering
 *  subcircuits at beginning, results on border at the bottom
 */
static void order_reverse()
{
 int node;
 ns.nm[0] = 0;
 for (node=1;  node<=STATUS::total_nodes;  ++node)
    ns.nm[node] = STATUS::total_nodes - node + 1;
}
/*--------------------------------------------------------------------------*/
/* order_forward: use user ordering, with subcircuits added to end
 * results in border at the top (worst possible if lots of subcircuits)
 */
static void order_forward()
{
 int node;
 ns.nm[0] = 0;
 for (node=1;  node<=STATUS::total_nodes;  ++node)
    ns.nm[node] = node;
}
/*--------------------------------------------------------------------------*/
/* order_auto: full automatic ordering
 * reverse, for now
 */
static void order_auto()
{
 int node;
 ns.nm[0] = 0;
 for (node=1;  node<=STATUS::total_nodes;  ++node)
    ns.nm[node] = STATUS::total_nodes - node + 1;
}
/*--------------------------------------------------------------------------*/
/* map_list: map intermediate node to internal (working) node number.
 * recursive: actual mapping done here
 */
static void map_list(CARD *stop)
{
 assert(exists(stop));
 CARD *brh;
 brh = stop;
 do {
    if (brh->isdevice()){
       for (int ii = 0;  brh->n[ii].t != INVALIDNODE;  ii++){
	  brh->n[ii].m = ns.nm[brh->n[ii].t];
       }
       brh->n[ii].m = INVALIDNODE;
       if (brh->sub()){
	  map_list(brh->sub());
       }
    }
 } while (brh = brh->next(),  brh != stop);
}
/*--------------------------------------------------------------------------*/
/* build_basenode_table: build a table containing the lowest node number
 * connected to each node.  This determines the bandwidth of the matrix,
 * or the length of the vectors (in the matrix) corresponding to each node.
 * Actual work is done by addto_basenode_table.
 */
/* addto_basenode_table: actual work of build_basenode_table.
 * Find and store "base-node" for each node.  This is the lowest numbered
 * node that it connects to.  We don't do anything with higher numbered nodes.
 * (Unnecessary because of symmetry.)   Recursive, for subckts.
 */
static void addto_basenode_table(const CARD *stop)
{
 assert(exists(stop));
 const CARD *brh;
 brh = stop;
 do {
    if (brh->isdevice()){
       int ii,jj;
       for (ii = 0;  brh->n[ii].m != INVALIDNODE;  ii++){
	  if (brh->n[ii].m != 0){
	     for (jj = 0;  jj < ii ;  jj++){
		aa.iwant(brh->n[ii].m,brh->n[jj].m);
		lu.iwant(brh->n[ii].m,brh->n[jj].m);
		acx.iwant(brh->n[ii].m,brh->n[jj].m);
	     }
	     ++nstat[brh->n[ii].m].needsanalog;
	  }
       }
       if (brh->sub()){
	  addto_basenode_table(brh->sub());
       }
    }
 } while (brh = brh->next(), brh != stop);
}
/*--------------------------------------------------------------------------*/
/* alloc_hold_vectors: allocate space to hold data between commands.
 * store voltage between commands, for restart and convergence assistance
 */
static void alloc_hold_vectors()
{
 nstat = new NODE[STATUS::total_nodes+2];
 if (!nstat)
    error(bERROR, e_om, "nstat");
 for (int ii=0;  ii <= STATUS::total_nodes;  ++ii){
    nstat[ns.nm[ii]].number = ii;
 }

 ns.vdc = (double*)calloc((unsigned)STATUS::total_nodes+2, sizeof(double));
 if (!ns.vdc)
    error(bERROR, e_om, "vdc");
}
/*--------------------------------------------------------------------------*/
/* alloc_constants: allocate constant vectors
 * (constant is not the right word...)  allocate space for the right-side
 * vector (initially current sources, on solution becomes voltages) and copies
 * used for one-time-ago and convergence checking
 */
static void alloc_constants()
{
 ns.ac = (complex*)calloc((unsigned)STATUS::total_nodes+2, sizeof(complex));
 if (!ns.ac)
    error(bERROR, e_om, "ac");

 ns.i = (double*)calloc((unsigned)STATUS::total_nodes+2, sizeof(double));
 if (!ns.i)
    error(bERROR, e_om, "i");

 ns.v0 = (double*)calloc((unsigned)STATUS::total_nodes+2, sizeof(double));
 if (!ns.v0)
    error(bERROR, e_om, "v0");

 ns.vt1 = (double*)calloc((unsigned)STATUS::total_nodes+2, sizeof(double));
 if (!ns.vt1)
    error(bERROR, e_om, "vt1");

 ns.fw = (double*)calloc((unsigned)STATUS::total_nodes+2, sizeof(double));
 if (!ns.fw)
    error(bERROR, e_om, "fw");
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
